//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#ifndef _ODEMANAGER_HXX_
#define _ODEMANAGER_HXX_

#include "function.hxx"
#include "double.hxx"
#include "string.hxx"
#include "list.hxx"
#include "mlist.hxx"
#include "pointer.hxx"
#include "struct.hxx"
#include "callable.hxx"
#include "runvisitor.hxx"
#include "context.hxx"

extern "C"
{
#include "localization.h"
#include "Scierror.h"
#include "sciprint.h"

void C2F(daxpy)(int*, double*, double*, int*, double*, int*);
void C2F(dgemv)(char*, int*, int*, double*, double*, int*, double*, int*, double*, double*, int*);
//void C2F(dgemm)(char*, char*, int*, int*, int*, double*, double*, int*, double*, int*, double*, double*, int*);
double C2F(ddot)(int*, double*, int*, double *, int*);

#define ODEMANAGER_KILLME(p) if (p != NULL) p->killMe();

/* Sundials includes */
#include <cvode/cvode_impl.h>
#include <cvode/cvode.h>            /* prototypes for CVODES fcts. and consts. */
#include <cvode/cvode_dense.h>     /* prototype for CVDense */
#include <cvode/cvode_direct.h>    /* prototypes for various DlsMat operations */
#include <nvector/nvector_serial.h>   /* serial N_Vector types, fcts., and macros */
#include <sundials/sundials_dense.h>  /* prototypes for various DlsMat operations */
#include <sundials/sundials_direct.h> /* definitions of DlsMat and DENSE_ELEM */
#include <sundials/sundials_types.h>  /* definition of type realtype */
#include <sundials/sundials_math.h>
}

typedef void(*dynlibFunPtr)();

class OdeManager
{
    public :

    OdeManager(const std::wstring& callerName, const std::wstring& solverName, OdeManager* prevManager = NULL)
    {
        m_wstrCaller = callerName;
        m_wstrSolver = solverName;
        m_prevManager = prevManager;
        m_iSizeOfCVODEInput[EVENTS] = -1;
        m_pCallFunction.fill(NULL);
        m_pEntryPointFunction.fill(NULL);
        m_pEntryPointFunctionWithPar.fill(NULL);
        m_dblOptT0 = NAN;
        m_dblInitialStep = NAN;
        m_iNbEvents = 0;
    }

    ~OdeManager() {
        if (m_cvode_mem != NULL) CVodeFree(&m_cvode_mem);
        if (m_N_VectorY != NULL) N_VDestroy_Serial(m_N_VectorY);
        if (m_N_VectorAtol != NULL) N_VDestroy_Serial(m_N_VectorAtol);
        //ODEMANAGER_KILLME(m_pDblTOut);
        //ODEMANAGER_KILLME(m_pDblYOut);
        ODEMANAGER_KILLME(m_pDblNSMatrices);
    };

    enum functionKind {RHS=0,JACY,EVENTS,INTCB};
    char pstrArgName[4][16] = {"f","\"jacobian\"","\"events\"","\"intcb\""};

    std::wstring getCallerName()
    {
        return m_wstrCaller;
    }
    
    types::Double *getTOut()
    {
        if (m_pDblTOut == NULL)
        {
            // time at the begining of each internal step of the method
            types::Double *pDblTOut = new types::Double(1,m_dblVecTOut.size());
            std::copy(m_dblVecTOut.begin(), m_dblVecTOut.end(), pDblTOut->get());
            return pDblTOut;
        }
        // user prescribed timesteps (eventually stoppped by event)
        return m_pDblTOut;
    }

    types::Double *getYOut()
    {
        if (m_pDblYOut == NULL)
        {
            // solution y at the begining of each internal step of the method
            types::Double *pDblYOut = createYOut(m_iNbEq, m_vecYOut.size());
            for (int i = 0; i < m_vecYOut.size(); i++)
            {
                double *pdbl = m_vecYOut[i].data();
                std::copy(pdbl, pdbl+m_iNbEq, pDblYOut->get() + i*m_iNbEq);
                if (m_odeIsComplex)
                {
                    std::copy(pdbl+m_iNbEq, pdbl+2*m_iNbEq, pDblYOut->getImg() + i*m_iNbEq);                
                }
            }
            return pDblYOut;            
        }
        // solution at user prescribed timesteps (eventually stoppped by event)
        return m_pDblYOut;
    }

    types::Double *getIndexEvent()
    {
        types::Double *pDbl = createYOut(m_iNbEvents, m_dblVecIndexEvent.size());
        for (int i = 0; i < m_dblVecIndexEvent.size(); i++)
        {
            std::copy(m_dblVecIndexEvent[i].begin(), m_dblVecIndexEvent[i].end(), pDbl->get() + i*m_iNbEvents);
        }
        return pDbl; 
    }

    types::Double *getTEvent()
    {
        types::Double *pDbl = new types::Double(1,m_dblVecEventTime.size());
        std::copy(m_dblVecEventTime.begin(), m_dblVecEventTime.end(), pDbl->get());
        return pDbl;
    }

    types::Double *getYEvent()
    {
        types::Double *pDblYEvent = createYOut(m_iNbEq, m_dblVecYEvent.size());
        for (int i = 0; i < m_dblVecYEvent.size(); i++)
        {
            double *pdbl = m_dblVecYEvent[i].data();
            std::copy(pdbl, pdbl+m_iNbEq, pDblYEvent->get() + i*m_iNbEq);
            if (m_odeIsComplex)
            {
                std::copy(pdbl+m_iNbEq, pdbl+2*m_iNbEq, pDblYEvent->getImg() + i*m_iNbEq);                
            }
        }
        return pDblYEvent;            
    }

    types::Double *getY0()
    {
        return m_pDblY0;
    }

    types::Double *getNordsieckMatrices()
    {
        return m_pDblNSMatrices;
    }

    std::vector<int> getNordsieckMatricesIndex()
    {
        return m_indexNS;
    }
    std::vector<double> getStepVector()
    {
        return m_dblVecStep;
    }
    std::vector<double> getCurrTimeVector()
    {
        return m_dblVecCurrTime;
    }
    std::vector<int> getPositive()
    {
        return m_iVecPositive;
    }

    int getNbEvents()
    {
        return m_iNbEvents;
    }
    
    void setIretCount (int iRetCount)
    {
        m_iRetCount = iRetCount;
    }
    
    int getNEq()
    {
        return m_iNbEq;
    }

    int getNRealEq()
    {
        return m_iNbRealEq;
    }
    
    void *getCVODE_mem()
    {
        return m_cvode_mem;
    }
    
    dynlibFunPtr getEntryPointFunction(functionKind  what)
    {
        return m_pEntryPointFunction[what];
    }

    dynlibFunPtr getEntryPointFunctionWithPar(functionKind  what)
    {
        return m_pEntryPointFunctionWithPar[what];
    }

    types::typed_list getParameters(functionKind  what)
    {
        return m_pParameters[what];
    }

    void parseFunction(types::InternalType *pIn, functionKind what);
    void parseMatrices(types::typed_list &in);
    void parseOptions(types::optional_list &opt);
    void setupEvents();
    void init();
    void solve();
    void computeFunction(double t, double *pdblY, double *pdblDY, functionKind what, double *pdblOut);
    bool intermediateCallback(double t, double *pdblY, wchar_t *pwstFlag);
    void createSolutionOutput(types::typed_list &out);

    template <typename T>
    types::Double *createAndExtendFrom(types::Double *pDblPrev, std::vector<T> &pVect);

    types::Double *createYOut(int iNbOut, int iSizeTSpan, bool bFlat = false);
    types::Struct *getStats();

    private :
    
    OdeManager* m_prevManager;
    
    void *m_cvode_mem = NULL;
    N_Vector m_N_VectorY = NULL;
    N_Vector m_N_VectorAtol = NULL;
    std::wstring m_wstrCaller;

    types::Callable *m_odeFun = NULL;
    types::Callable *m_odeJac= NULL;

    types::Double *m_pDblTSpan = NULL;    
    types::Double *m_pDblY0 = NULL;    
    types::Double *m_pDblTOut = NULL;
    types::Double *m_pDblYOut = NULL;
    types::Double *m_pDblNSMatrices = NULL;

    std::vector<int> m_indexNS;    
    std::vector<double> m_dblVecAtol;
    std::vector<double> m_dblVecTOut;
    std::vector<double> m_dblVecStep;
    std::vector<double>  m_dblVecCurrTime;
    std::vector<int> m_iVecOrder;
    std::vector<int> m_iVecJacBand;
    std::vector<std::vector<double>> m_vecYOut;
    
    std::vector<int> m_iVecEventIsTerminal;
    std::vector<int> m_iVecEventDirection;

    std::vector<std::vector<double>> m_dblVecYEvent;
    std::vector<std::vector<double>> m_dblVecIndexEvent;
    std::vector<double> m_dblVecEventTime;
    std::vector<int> m_iVecPositive;
    
    bool m_odeIsComplex = false;
    
    double m_dblT0;
    double m_dblOptT0;
    double m_dblRtol;
    double m_dblAtol;
    double m_dblCurrentStep;
    double m_dblInitialStep;  
    double m_dblMaxStep;  
    double m_dblMinStep;  
    int m_iRetCount;
    int m_iNbEq;
    int m_iNbRealEq;
    int m_iMaxOrder;
    int m_iNbEvents;
    int m_iNRefine;

    long int m_incStat[9] = {0,0,0,0,0,0,0,0,0};

    std::wstring m_wstrMethod;
    std::wstring m_wstrSolver;
    
    types::typed_list m_odeFunParameters;
    types::typed_list m_odeJacParameters;

    std::array<dynlibFunPtr,5> m_pEntryPointFunction;
    std::array<dynlibFunPtr,5> m_pEntryPointFunctionWithPar;
    std::array<types::Callable *,5> m_pCallFunction;
    std::array<char *,5> m_pCallFunctionName;
    std::array<int,5> m_iSizeOfCVODEInput;
    std::array<types::typed_list,5> m_pParameters;    
};

#endif
