//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "CVODEBridge.hxx"

static std::map<std::wstring, int> intCbState{{L"init",-1}, {L"step",0}, {L"done",1}};

#define MSGCV_HNIL_STOP "Internal " MSG_TIME_H " are such that t + h = t on the next step. Singularity likely."

int CVODEFun(realtype t, N_Vector N_VectorY, N_Vector N_VectorYDot, void *pManager)
{
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    // positiveness constraints
    for (auto i : manager->getPositive())
    {
        if (NV_Ith_S(N_VectorY,i-1) < 0)
        {
            return 1; // return a CVODE recoverable error
        }
    }

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::RHS);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::RHS);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        ((CVODEDynFun)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYDot));
    }
    else if (pFuncWithPar != NULL)
    {
        types::Double *pDblPar = manager->getParameters(OdeManager::RHS)[0]->getAs<types::Double>();
        ((CVODEDynFunWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorYDot), pDblPar->get());
    }
    else
    {
        manager->computeFunction(t, NV_DATA_S(N_VectorY), NULL, OdeManager::RHS, NV_DATA_S(N_VectorYDot));
    }
    
    return 0;
}

int CVODEJac(long int iNEq, realtype t, N_Vector N_VectorY, N_Vector N_VectorFy, DlsMat DlsMat_J, void *pManager, N_Vector tmp1, N_Vector tmp2, N_Vector tmp3)
{
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::JACY);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::JACY);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        ((CVODEDynJac)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorFy), DlsMat_J->data);
    }
    else if (pFuncWithPar != NULL)
    {
        int iNbRealEq = manager->getNRealEq();
        types::Double *pDblPar = manager->getParameters(OdeManager::RHS)[0]->getAs<types::Double>();
        ((CVODEDynJacWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), NV_DATA_S(N_VectorFy), DlsMat_J->data, pDblPar->get());
    }
    else
    {
        manager->computeFunction(t, NV_DATA_S(N_VectorY), NULL, OdeManager::JACY, DlsMat_J->data);        
    }
    return 0;
}

int CVODEEvent(realtype t, N_Vector N_VectorY, realtype *pdblOut, void *pManager)
{
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::EVENTS);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::EVENTS);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        int iNbEvents = manager->getNbEvents();
        ((CVODEDynEvent)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iNbEvents, pdblOut, NULL, NULL);
    }
    else if (pFuncWithPar != NULL)
    {
        int iNbEvents = manager->getNbEvents();
        types::Double *pDblPar = manager->getParameters(OdeManager::EVENTS)[0]->getAs<types::Double>();
        ((CVODEDynEventWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iNbEvents, pdblOut, NULL, NULL, pDblPar->get());        
    }
    else
    {
        manager->computeFunction(t, NV_DATA_S(N_VectorY), NULL, OdeManager::EVENTS, pdblOut);        
    }
    return 0;
}

int CVODEIntCb(realtype t, N_Vector N_VectorY, wchar_t *pwstFlag, void *pManager)
{
    int iRet = 0;
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    int iNbRealEq = manager->getNRealEq();

    dynlibFunPtr pFunc = manager->getEntryPointFunction(OdeManager::INTCB);
    dynlibFunPtr pFuncWithPar = manager->getEntryPointFunctionWithPar(OdeManager::INTCB);
    if (pFunc != NULL)
    {
        // Computation is made by user C/C++/Fortran dynamically lindked entrypoint
        int iState = intCbState[pwstFlag];
        iRet = ((CVODEDynIntCb)pFunc)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iState);
    }
    else if (pFuncWithPar != NULL)
    {
        types::Double *pDblPar = manager->getParameters(OdeManager::INTCB)[0]->getAs<types::Double>();
        int iState = intCbState[pwstFlag];
        iRet = ((CVODEDynIntCbWithPar)pFuncWithPar)(&iNbRealEq, &t, NV_DATA_S(N_VectorY), &iState, pDblPar->get());
    }
    else
    {
        iRet = (int) manager->intermediateCallback(t, NV_DATA_S(N_VectorY), pwstFlag);
    }
    
    return iRet;
}


void CVODEErrHandler(int error_code, const char *module, const char *function, char *msg, void *pManager)
{
    char errorMsg[256];
    OdeManager *manager = static_cast<OdeManager *>(pManager);
    if (msg == NULL)
    {
        char cvodeMsg[256] = "";
        double dblTime, dblStep;
        CVodeGetCurrentTime(manager->getCVODE_mem(), &dblTime);
        CVodeGetLastStep(manager->getCVODE_mem(), &dblStep);
        /* Depending on flag, print error message and return error flag */
        switch (error_code)
        {
            case CV_CONV_FAILURE:
                sprintf(cvodeMsg, MSGCV_CONV_FAILS, dblTime, dblStep);
                break;
            case CV_TOO_MUCH_WORK:
                sprintf(cvodeMsg, MSGCV_MAX_STEPS, dblTime);
                break;
            case CV_REPTD_RHSFUNC_ERR:
                sprintf(cvodeMsg, MSGCV_RHSFUNC_REPTD, dblTime);
                break;          
            case CV_WARNING:
                sprintf(cvodeMsg, MSGCV_HNIL_STOP, dblTime, dblStep);
                break;
            default:
                sprintf(cvodeMsg,"%s at t=%g\n", CVodeGetReturnFlagName((long int) error_code), dblTime);
                break; 
        }        
        sprintf(errorMsg, _("%ls: %s\n"), manager->getCallerName().c_str(), cvodeMsg);
    }
    else
    {
        sprintf(errorMsg, _("%ls: %s \n"), manager->getCallerName().c_str(), msg);
    }        
    if (error_code != CV_WARNING)
    {
        throw ast::InternalError(errorMsg);        
    }
    else
    {
        sciprint("\n%s",errorMsg);
    }
}
