//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"

types::Struct *OdeManager::getStats()
{
    double dblStat[6] = {0.0,0.0,0.0,0.0,0.0,0.0};
    int qlast;
    int qcur;

    std::wstring fieldNames[14] = {L"nsteps", L"nfevals", L"nfevalsFD", L"njacevals", L"neventevals",
    L"nlinsolve", L"nrejsteps", L"nnliters", L"nnlcvfails", L"order", L"h0", L"hlast",
    L"hcur", L"tcur"};
    
    if (m_pCallFunction[JACY] != NULL)
    {
        m_incStat[2] = 0;
    }
        
    CVodeGetNonlinSolvStats(m_cvode_mem, m_incStat+7, m_incStat+8);
    CVodeGetIntegratorStats(m_cvode_mem, m_incStat, m_incStat+1, m_incStat+5, m_incStat+6, &qlast, &qcur, dblStat, dblStat+1, dblStat+2, dblStat+3);
    CVDlsGetNumRhsEvals(m_cvode_mem, m_incStat+2);
    CVDlsGetNumJacEvals(m_cvode_mem, m_incStat+3);
    CVodeGetNumGEvals(m_cvode_mem, m_incStat+4);
    CVodeGetNumLinSolvSetups(m_cvode_mem,m_incStat+5); // CVodeGetIntegratorStats seems broken for nlinsolve(m_incStat+5), hence we do a direct call
    
    // dblStat[0] = (double)qlast;
    // dblStat[1] = (double)qcur;

    // if extending a previous solution, update incremental stats only
    if (m_prevManager != NULL)
    {
        for (int i=0; i<9; i++)
        {
           m_incStat[i] += m_prevManager-> m_incStat[i]; 
        }
    }
    
    types::Struct *pSt = new types::Struct(1,1);
    for (int i=0; i<9; i++)
    {
        pSt->addField(fieldNames[i].c_str());        
        pSt->get(0)->set(fieldNames[i].c_str(),new types::Double((double)m_incStat[i]));        
    }

    // order of method for each step
    types::Double *pDblOrder = new types::Double(1,m_iVecOrder.size());
    std::copy(m_iVecOrder.begin(), m_iVecOrder.end(), pDblOrder->get());
    pSt->addField(fieldNames[9].c_str());        
    pSt->get(0)->set(fieldNames[9].c_str(), pDblOrder);        
    
    for (int i=10; i<14; i++)
    {
        pSt->addField(fieldNames[i].c_str());        
        pSt->get(0)->set(fieldNames[i].c_str(), new types::Double(dblStat[i-10]));        
    }
    return pSt;
}

types::Double *OdeManager::createYOut(int iNbOut, int iSizeTSpan, bool bFlat)
{
    types::Double *pDblYOut;
    int iDimsY0 = m_pDblY0->getDims();

    if (bFlat == false && (iDimsY0 > 2 || (m_pDblY0->getCols() > 1 & iDimsY0 == 2)))
    {
        int *piDimsArrayY0 = m_pDblY0->getDimsArray();
        int *piDimsArrayYOut = new int(iDimsY0+1);
        for (int i=0; i<iDimsY0; i++)
        {
            piDimsArrayYOut[i] = piDimsArrayY0[i];
        }
        piDimsArrayYOut[iDimsY0] = iSizeTSpan;
        pDblYOut = new types::Double(iDimsY0+1, piDimsArrayYOut, m_odeIsComplex);
    }
    else
    {
        pDblYOut = new types::Double(iNbOut, iSizeTSpan, m_odeIsComplex);
    }
    return pDblYOut;
}

void OdeManager::createSolutionOutput(types::typed_list &out)
{
    // return a MList of "_odeSolution" type with method solution at internal steps
    // and a pointer to the OdeManager object. Extraction on this MList calls %_odeSolution_e and allows
    // to compute solution at arbitrary value of time by using method dense interpolant.
    bool bEvents = m_iNbEvents > 0;
    types::MList *pObj = new types::MList();
    types::String *pStr = new types::String(1,7+((int)bEvents)*3);
    int k=0;
    
    pStr->set(k++,L"_odeSolution");
    pStr->set(k++,L"solver");
    pStr->set(k++,L"method");
    pStr->set(k++,L"t");
    pStr->set(k++,L"y");
    if (bEvents == true)
    {
        pStr->set(k++,L"te");
        pStr->set(k++,L"ye");            
        pStr->set(k++,L"ie");            
    }
    pStr->set(k++,L"stats");
    pStr->set(k++,L"idata");

    k = 0;
    pObj->set(k++,pStr);
    pObj->set(k++,new types::String(m_wstrSolver.c_str()));
    pObj->set(k++,new types::String(m_wstrMethod.c_str()));
    pObj->set(k++,getTOut());
    pObj->set(k++,getYOut());
    if (bEvents == true)
    {
        pObj->set(k++, getTEvent());
        pObj->set(k++, getYEvent());            
        pObj->set(k++, getIndexEvent());            
    }
    pObj->set(k++, getStats());
    pObj->set(k++,new types::Pointer((void *)this));

    out.push_back(pObj); 
}
