//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "CVODEBridge.hxx"

void OdeManager::init()
{
    char errorMsg[256];

    if (m_wstrMethod == L"adams")
    {
        m_cvode_mem = CVodeCreate(CV_ADAMS, CV_FUNCTIONAL);
        if (m_cvode_mem == NULL) 
        {
            sprintf(errorMsg,"CVodeCreate error\n");
            throw ast::InternalError(errorMsg);
        }
    }
    else if ( m_wstrMethod == L"bdf")
    {
        m_cvode_mem = CVodeCreate(CV_BDF, CV_NEWTON);
        if (m_cvode_mem == NULL) 
        {
            sprintf(errorMsg,"CVodeCreate error\n");
            throw ast::InternalError(errorMsg);
        }
    }

    if (std::isnan(m_dblOptT0))
    {
        if (m_pDblTSpan->getSize() > 1)
        {
            m_dblT0 = m_pDblTSpan->get(0);   
        }
        else
        {
            sprintf(errorMsg, _("%ls: parameter \"%ls\" must be defined by user.\n"), m_wstrCaller.c_str(), L"t0");
            throw ast::InternalError(errorMsg);
        } 
    }
    else
    {
        m_dblT0 = m_dblOptT0;
    }
    
    // Complex ode is handled by using [real(y); imag(y)] as state, 
    // hence the number of *real* equations is twice the number of complex equatioms
    m_iNbRealEq = m_odeIsComplex ? 2*m_iNbEq : m_iNbEq;

    // Load Y0 into N_Serial vector
    // When ODE is complex  m_N_VectorY is the concatenation of real and imaginary part of user Y
    m_N_VectorY = N_VNew_Serial(m_iNbRealEq);

    std::copy(m_pDblY0->get(), m_pDblY0->get()+m_iNbEq, NV_DATA_S(m_N_VectorY));
    if (m_odeIsComplex)
    {
        std::copy(m_pDblY0->getImg(), m_pDblY0->getImg()+m_iNbEq, NV_DATA_S(m_N_VectorY)+m_iNbEq);        
    }
    
    if (CVodeInit(m_cvode_mem, CVODEFun, m_dblT0, m_N_VectorY) != CV_SUCCESS)
    {
        sprintf(errorMsg,"CVodeInit error\n");
        throw ast::InternalError(errorMsg);
    }
    
    if (CVodeSetUserData(m_cvode_mem, (void *)(this)) != CV_SUCCESS)
    {
        sprintf(errorMsg,"CVodeSetUserData error\n");
        throw ast::InternalError(errorMsg);
    }

    if (m_dblInitialStep > 0)
    {
        if (CVodeSetInitStep(m_cvode_mem, m_dblInitialStep) != CV_SUCCESS)
        {
            sprintf(errorMsg,"CVodeSetInitStep error\n");
            throw ast::InternalError(errorMsg);
        }
    }

    if (m_dblMaxStep > 0)
    {
        if (CVodeSetMaxStep(m_cvode_mem, m_dblMaxStep) != CV_SUCCESS)
        {
            sprintf(errorMsg,"CVodeSetMaxStep error\n");
            throw ast::InternalError(errorMsg);
        }
    }

    if (m_dblMinStep > 0)
    {
        if (CVodeSetMinStep(m_cvode_mem, m_dblMinStep) != CV_SUCCESS)
        {
            sprintf(errorMsg,"CVodeSetMinStep error\n");
            throw ast::InternalError(errorMsg);
        }
    }

    // Events
    if (m_iNbEvents > 0)
    {        
        if (CVodeRootInit(m_cvode_mem, m_iNbEvents, CVODEEvent) != CV_SUCCESS)
        {
            sprintf(errorMsg,"CVodeRootInit error\n");
            throw ast::InternalError(errorMsg); 
        }
        if (m_iVecEventDirection.size() > 0)
        {
            if (CVodeSetRootDirection(m_cvode_mem, m_iVecEventDirection.data()) != CV_SUCCESS)
            {
                sprintf(errorMsg,"CVodeRootInit error\n");
                throw ast::InternalError(errorMsg);
            }
        }
     }

    // Absolute (a vector) and relative tolerance
    m_N_VectorAtol = N_VNew_Serial(m_iNbRealEq);
    if (m_odeIsComplex)
    {
        m_dblVecAtol.resize(m_iNbRealEq);
        for (int i=0; i<m_iNbEq; i++)
        {
            m_dblVecAtol[i+m_iNbEq] = m_dblVecAtol[i];
        }
    }
    std::copy(m_dblVecAtol.begin(), m_dblVecAtol.end(), NV_DATA_S(m_N_VectorAtol));
    if (CVodeSVtolerances(m_cvode_mem, m_dblRtol, m_N_VectorAtol) != CV_SUCCESS)
    {
        sprintf(errorMsg,"CVodeSVtolerances error\n");
        throw ast::InternalError(errorMsg);
    }

    if (CVodeSetMaxOrd(m_cvode_mem, m_iMaxOrder) != CV_SUCCESS)
    {
        sprintf(errorMsg,"CVodeSetMaxOrder error\n");
        throw ast::InternalError(errorMsg);
    }

    if (m_wstrMethod == L"bdf")
    {
        if (CVDense(m_cvode_mem, m_iNbRealEq) != CV_SUCCESS)
        {
            sprintf(errorMsg,"CVDense error\n");
            throw ast::InternalError(errorMsg);
        }

        if (m_pCallFunction[JACY] != NULL)
        {   
            if (CVDlsSetDenseJacFn(m_cvode_mem, CVODEJac) != CV_SUCCESS)
            {
                sprintf(errorMsg,"CVDlsSetDenseJacFn error\n");
                throw ast::InternalError(errorMsg);
            }
        }
        else
        {
            if (CVDlsSetDenseJacFn(m_cvode_mem, NULL) != CV_SUCCESS)
            {
                sprintf(errorMsg,"CVDlsSetDenseJacFn error\n");
                throw ast::InternalError(errorMsg);
            }                    
        }
        CVodeSetErrHandlerFn(m_cvode_mem, CVODEErrHandler, (void *)(this));
    }
}