//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "function.hxx"
#include "double.hxx"
#include "string.hxx"
#include "list.hxx"
#include "mlist.hxx"
#include "pointer.hxx"
#include "struct.hxx"
#include "callable.hxx"
#include "runvisitor.hxx"
#include "context.hxx"

#include "OdeManager.hxx"
#include "odeparameters.hxx"

extern "C"
{
#include "localization.h"
#include "Scierror.h"
#include "sciprint.h"
}

types::Function::ReturnValue sci_cvode_extend(types::typed_list &in, types::optional_list &opt, int _iRetCount, types::typed_list &out)
{
    OdeManager *manager = NULL;
    OdeManager *prevManager = NULL;
    char errorMsg[256];

    if (in.size() != 2)
    {
        Scierror(999, _("%s: Wrong number of input argument(s): 2 expected.\n"), "cvode_extend", 2);
        return types::Function::Error;        
    }
    if (_iRetCount > 1)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d expected.\n"), "cvode_extend", 1);            
        throw ast::InternalError(errorMsg);          
    }
    if (in[0]->isMList() == false || in[0]->getAs<types::MList>()->getTypeStr() != L"_odeSolution")
    {
        sprintf(errorMsg, _("%s: Wrong type of input argument #%d: %s expected.\n"), "cvode_extend", 1, "_odeSolution");            
        throw ast::InternalError(errorMsg);          
    }
    types::InternalType *pI;
    types::MList *pObj = in[0]->getAs<types::MList>();
    if (pObj->extract(L"idata",pI) && pI->isPointer())
    {
        prevManager = static_cast<OdeManager *>(pI->getAs<types::Pointer>()->get());
    }

    manager = new OdeManager(L"cvode_extend", L"cvode", prevManager);        

    manager->setIretCount(1);

    try
    {
        manager->parseMatrices(in);
        manager->parseOptions(opt);
        manager->init();            
        manager->solve();
    }
    catch(ast::InternalError& ie)
    {
        delete manager;
        return types::Function::Error;        
    }
    catch(ast::InternalAbort& ia)
    {
        delete manager;        
        throw(ia);
    }
    
    manager->createSolutionOutput(out);

    return types::Function::OK;
}

types::Function::ReturnValue sci_cvode_solve(types::typed_list &in, types::optional_list &opt, int _iRetCount, types::typed_list &out)
{
    OdeManager *manager = NULL;
    OdeManager *prevManager = NULL;
    char errorMsg[256];
    int iStart=0;
    
    if (in.size() != 3)
    {
        Scierror(999, _("%s: Wrong number of input argument(s): %d expected.\n"), "cvode_solve", 3);
        return types::Function::Error;        
    }
    if (_iRetCount > 5)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d to %d expected.\n"), "cvode_solve", 1, 5);            
        throw ast::InternalError(errorMsg);          
    }
    
    manager = new OdeManager(L"cvode_solve", L"cvode");

    manager->setIretCount(_iRetCount);

    try
    {
        manager->parseMatrices(in);
        manager->parseFunction(in[iStart], OdeManager::RHS);
        manager->parseOptions(opt);
        manager->init();
        manager->solve();
    }
    catch(ast::InternalError& ie)
    {
        delete manager;
        return types::Function::Error;
    }
    catch(ast::InternalAbort& ia)
    {
        delete manager;
        throw(ia);
    }

    if (_iRetCount <= 1)
    {
        manager->createSolutionOutput(out);
    }
    else
    {
        out.push_back(manager->getTOut());
        out.push_back(manager->getYOut());
        if (manager->getNbEvents() > 0)
        {
            out.push_back(manager->getTEvent());
            out.push_back(manager->getYEvent());            
            out.push_back(manager->getIndexEvent());                            
        }
        delete manager;
    }

    return types::Function::OK;
}

types::Function::ReturnValue sci_percent_odeSolution_clear(types::typed_list &in, int _iRetCount, types::typed_list &out)
{
    OdeManager *manager = NULL;
    char errorMsg[256];

    if (in.size() == 1)
    {
        if (in[0]->isMList())
        {
            types::InternalType *pI;
            types::MList *pObj = in[0]->getAs<types::MList>();
            if (pObj->extract(L"idata",pI) && pI->isPointer())
            {
                manager = static_cast<OdeManager *>(pI->getAs<types::Pointer>()->get());
                if (manager != NULL)
                {
                    delete manager;
                }
            }
            else
            {
                sprintf(errorMsg, _("%s: Wrong type for argument #1.\n"), "%_odeSoutionc_clear");
                throw ast::InternalError(errorMsg);
            }
        }
        else
        {
            sprintf(errorMsg, _("%s: Wrong type for argument #1.\n"), "%_odeSoutionc_clear");
            throw ast::InternalError(errorMsg);
        }
    }
    else
    {
        sprintf(errorMsg, _("%s: Wrong number of input arguments.\n"), "%_odeSoutionc_clear");
        throw ast::InternalError(errorMsg);
    }

    return types::Function::OK;
}

types::Function::ReturnValue sci_percent_odeSolution_e(types::typed_list &in, int _iRetCount, types::typed_list &out)
{
    OdeManager *manager = NULL;
    types::Double *pDblYOut = NULL;
    types::Double *pDblYDOut = NULL;
    types::Double *pDblIndex = NULL;
    types::InternalType *pIn = NULL;
    
    double dblOne = 1.0; 
    double dblZero = 0.0; 
    double pdblCoeff[13];
    double *pdblYOut;
    double *pdblYOutImg;
    double *pdblYDOut;
    double *pdblYDOutImg;
    int iNEq;
    int iNRealEq;
    int iNEqOut;

    int iOne = 1;
    char errorMsg[256];
    
    if (in.size() >= 2)
    {
        if (in[in.size()-1]->isMList())
        {
            types::InternalType *pI;
            types::MList *pObj = in[in.size()-1]->getAs<types::MList>();
            if (pObj->extract(L"idata",pI) && pI->isPointer())
            {
                manager = static_cast<OdeManager *>(pI->getAs<types::Pointer>()->get());
                iNEq = manager->getNEq();
                iNRealEq = manager->getNRealEq();
                iNEqOut = iNEq;
                in.pop_back();
                
                if (in.size() == 2)
                {
                    pIn = in[1];
                    if (in[1]->isImplicitList())
                    {
                        types::ImplicitList *pIL = in[1]->getAs<types::ImplicitList>();
                        if (pIL->isComputable())
                        {
                            pIn = pIL->extractFullMatrix();
                        }
                    }
                    
                    if (pIn->isDouble() && pIn->getAs<types::Double>()->isComplex() == false)
                    {
                        pDblIndex = pIn->getAs<types::Double>();
                        for (int i=0; i < pDblIndex->getSize(); i++)
                        {
                            if (pDblIndex->get(i) != std::floor(pDblIndex->get(i)) || pDblIndex->get(i) < 1 || pDblIndex->get(i) > iNEq)
                            {
                                Scierror(999, _("%s: Wrong value for input argument #%d: must be an integer between %d and %d.\n"), "%_odeSolution_e",2, 1, iNEq);
                                return types::Function::Error;
                            }
                        }
                        iNEqOut = pDblIndex->getSize();
                    }
                    else
                    {
                        Scierror(999, _("%s: Wrong value for input argument #%d: must be an integer between %d and %d.\n"), "%_odeSolution_e",2, 1,iNEq);
                        return types::Function::Error;
                    }
                }

                pIn = in[0];
                if (in[0]->isImplicitList())
                {
                    types::ImplicitList *pIL = in[0]->getAs<types::ImplicitList>();
                    if (pIL->isComputable())
                    {
                        pIn = pIL->extractFullMatrix();
                    }
                }
                
                if (pIn->isDouble() && pIn->getAs<types::Double>()->isComplex() == false)
                {
                    types::Double *pDblUserTSpan = pIn->getAs<types::Double>();
                    double *pdblUserTSpan = pDblUserTSpan->get();
                    double *pdblTout = manager->getTOut()->get();
                    int iSizeTout = manager->getTOut()->getSize();
                    int iSizeUserTSpan = pDblUserTSpan->getSize();
                    
                    if (*std::min_element(pdblUserTSpan, pdblUserTSpan+iSizeUserTSpan) < pdblTout[0]
                        || *std::max_element(pdblUserTSpan, pdblUserTSpan+iSizeUserTSpan) > pdblTout[iSizeTout-1])
                    {
                        Scierror(999, _("%s: solution cannot be evaluated outside the interval [%g,%g].\n"), "%_odeSolution_e",pdblTout[0],pdblTout[iSizeTout-1]);
                        return types::Function::Error;        
                    }
                                        
                    // int *piDerDims = manager->getNordsieckMatrices()->getDimsArray();
                    // int iNSMatrixSize = piDerDims[0]*piDerDims[1];
                    // int iMaxOrderPlusOne = std::min(13,piDerDims[1]);
                    // int iMaxOrder = iMaxOrderPlusOne-1;
                    
                    if (iSizeTout > 1) // non degenerate case
                    {
                        pDblYOut = manager->createYOut(iNEqOut, iSizeUserTSpan, in.size() == 2);
                    
                        int iDims = pDblYOut->getDims();
                        int *piDims = pDblYOut->getDimsArray();
                    
                        pdblYOut = pDblYOut->get();
                        pdblYOutImg = pDblYOut->getImg();
                        if (_iRetCount == 2) // derivative requested
                        {
                            pDblYDOut = manager->createYOut(iNEqOut, iSizeUserTSpan, in.size() == 2);
                            pdblYDOut = pDblYDOut->get();                            
                            pdblYDOutImg = pDblYDOut->getImg();                            
                        }
                        std::vector<double> dblSolverTime = manager->getCurrTimeVector();
                        std::vector<int> piIndexNS = manager->getNordsieckMatricesIndex();
                        for (int i = 0; i < iSizeUserTSpan; i++)
                        {
                            //  locate user time values with method steps

                            double dblTUser = pDblUserTSpan->get(i);
                            double *pdblTime = std::lower_bound(pdblTout+1, pdblTout+iSizeTout, dblTUser);
                            int iIndex = pdblTime-pdblTout;
                            
                            // here (pdblTime[0] - pdblTime[-1]) and dblSolverTime[iIndex-1] - pdblTime[-1] differ when
                            // when pdblTime[0] is time of an event (we need actual solver step which is always larger).
                            
                            double dblt0 = dblSolverTime[iIndex-1];
                            double dblInvStep = 1.0/(dblt0-pdblTime[-1]);
                            double dblS = (dblTUser-dblt0)/(dblt0-pdblTime[-1]);
                            double dblCoeff = 1.0;
                            
                            double *pdblNS = manager->getNordsieckMatrices()->get() + piIndexNS[iIndex-1]*iNRealEq;
                            int iMethodOrderPlusOne =  piIndexNS[iIndex] - piIndexNS[iIndex-1];
                            int iMethodOrder = iMethodOrderPlusOne-1;
                            //  evaluation of interpolating polynomial is done by a matrix/vector multiplication
                            //  pdblCoeff = {1,s,s^2,...,s^q} where q is method order

                            for (int k = 0; k < iMethodOrderPlusOne; k++, dblCoeff *= dblS)
                            {
                                pdblCoeff[k] = dblCoeff;
                            }
                            if (in.size() == 1)
                            {
                                C2F(dgemv)("N", &iNEq, &iMethodOrderPlusOne, &dblOne, pdblNS, &iNRealEq, pdblCoeff, &iOne, &dblZero, pdblYOut, &iOne);
                                if (pdblYOutImg != NULL)
                                {
                                    // Imaginary part of Nordsieck matrix is in lower [iNEq x iMaxOrderPlusOne] block
                                    C2F(dgemv)("N", &iNEq, &iMethodOrderPlusOne, &dblOne, pdblNS+iNEq, &iNRealEq, pdblCoeff, &iOne, &dblZero, pdblYOutImg, &iOne);
                                    pdblYOutImg += iNEqOut;
                                }
                            }
                            else
                            {
                                // Compute only selected components
                                for (int j=0; j < iNEqOut; j++)
                                {
                                    double *pdblStartRow = pdblNS + (int)(pDblIndex->get(j)-1);
                                    pdblYOut[j] = C2F(ddot)(&iMethodOrderPlusOne, pdblStartRow, &iNRealEq, pdblCoeff, &iOne);
                                }
                                if (pdblYOutImg != NULL)
                                {
                                    for (int j=0; j < iNEqOut; j++)
                                    {
                                        // Imaginary part of Nordsieck matrix is in lower [iNEq x iMaxOrderPlusOne] block 
                                        double *pdblStartRow = pdblNS + (int)(pDblIndex->get(j)-1) + iNEq;
                                        pdblYOutImg[j] = C2F(ddot)(&iMethodOrderPlusOne, pdblStartRow, &iNRealEq, pdblCoeff, &iOne);
                                    }
                                    pdblYOutImg += iNEqOut;
                                }
                            }
                            pdblYOut += iNEqOut;
                        
                            if (_iRetCount == 2)
                            {
                                // derivative computation
                                // Skip first column of Nordsieck matrix and multiply/truncate pdblCoeff = {1,2*s,3*s^2,...,q*s^(q-1)}/h
                                for (int k = 1; k < iMethodOrderPlusOne; k++)
                                {
                                    pdblCoeff[k] *= k+1;
                                }
                                if (in.size() == 1)
                                {
                                    // Compute all components
                                    C2F(dgemv)("N", &iNEq, &iMethodOrder, &dblInvStep, pdblNS+iNRealEq, &iNRealEq, pdblCoeff, &iOne, &dblZero, pdblYDOut, &iOne);
                                    if (pdblYDOutImg != NULL)
                                    {
                                        C2F(dgemv)("N", &iNEq, &iMethodOrder, &dblInvStep, pdblNS+iNRealEq+iNEq, &iNRealEq, pdblCoeff, &iOne, &dblZero, pdblYDOutImg, &iOne);                                        
                                        pdblYDOutImg += iNEqOut;                                                                
                                    }
                                }
                                else
                                {
                                    // Compute only selected components
                                    for (int j=0; j < iNEqOut; j++)
                                    {
                                        double *pdblStartRow = pdblNS + iNRealEq + (int)(pDblIndex->get(j)-1);
                                        pdblYDOut[j] = C2F(ddot)(&iMethodOrder, pdblStartRow, &iNRealEq, pdblCoeff, &iOne)*dblInvStep;
                                    }
                                    if (pdblYDOutImg != NULL)
                                    {
                                        for (int j=0; j < iNEqOut; j++)
                                        {
                                            double *pdblStartRow = pdblNS + iNRealEq+iNEq + (int)(pDblIndex->get(j)-1);
                                            pdblYDOutImg[j] = C2F(ddot)(&iMethodOrder, pdblStartRow, &iNRealEq, pdblCoeff, &iOne)*dblInvStep;                                                                                    
                                        }
                                        pdblYDOutImg += iNEqOut;                                                                
                                    }
                                }
                                pdblYDOut += iNEqOut;                                                                
                            }
                        }                        
                    }
                    else
                    { // degenerate case, yield Y0 and eventualy f(T0,Y0)
                        if (in.size() == 1)
                        {
                            pDblYOut = manager->getY0()->clone();
                        }
                        else
                        { // return only selected components
                            types::Double *pDblY0 = manager->getY0();
                            pDblYOut = new types::Double(iNEqOut, 1, pDblY0->isComplex());
                            for (int j=0; j < iNEqOut; j++)
                            {
                                pDblYOut->set(j,pDblY0->get((int)(pDblIndex->get(j)-1)));
                                if (pDblYOut->isComplex())
                                {
                                    pDblYOut->setImg(j,pDblY0->getImg((int)(pDblIndex->get(j)-1)));
                                }
                            }                                                            
                        }
                        if (_iRetCount == 2)
                        {
                            // if (in.size() == 1)
                            // {
                            //     manager->computeFunction(pdblUserTSpan[0], manager->getY0()->get(), NULL, OdeManager::RHS, pdblYDOut);
                            // }
                            // else
                            // { // return only selected components
                            //     double *pdblF0 = new double(iNEq);
                            //     manager->computeFunction(pdblUserTSpan[0], manager->getY0()->get(), NULL, OdeManager::RHS, pdblF0);
                            //     for (int j=0; j < iNEqOut; j++)
                            //     {
                            //         pdblYDOut[j] = pdblF0[(int)(pDblIndex->get(j)-1)];
                            //     }
                            //     delete pdblF0;
                            // }
                            pDblYDOut = types::Double::Empty();
                        }
                    }
                    
                    out.push_back(pDblYOut);
                    if (_iRetCount == 2)
                    {
                        out.push_back(pDblYDOut);                            
                    }
                }
                else
                {
                    Scierror(999, _("%s: Wrong type for input argument #%d: A real matrix expected.\n"), "_odeSolution_e",1);
                    return types::Function::Error;        
                }                    
                
                return types::Function::OK;                
            }
            else
            {
                Scierror(999,_("%s: Wrong type for argument #1.\n"), "%_odeSolution_e");
                return types::Function::Error;
            }
        }
        else
        {
            Scierror(999,_("%s: Wrong type for argument #1.\n"), "%_odeSolution_e");
            return types::Function::Error;
        }
    }
    else
    {
         Scierror(999, _("%s: Wrong number of input arguments.\n"), "%_odeSolution_e");
         return types::Function::Error;
    }
    return types::Function::OK;
}

















