//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "odeparameters.hxx"

void OdeManager::parseFunction(types::InternalType *pIn, functionKind what)
{
    char errorMsg[256];
    
    if (pIn == NULL)
    {
        // we are extending a solution with same Rhs, Jacobian or Events function
        m_pCallFunction[what] = m_prevManager-> m_pCallFunction[what]; // can be NULL
        m_pParameters[what] = m_prevManager->m_pParameters[what]; // can be NULL
    }
    else if (pIn->isCallable())
    {
        m_pCallFunction[what] = pIn->getAs<types::Callable>();
    }
    else if (pIn->isList())
    {
        types::List *pList = pIn->getAs<types::List>();
        
        types::InternalType *pIFirst = pList->getSize() > 0 ? pList->get(0) : NULL;
        if (pIFirst != NULL && pIFirst->isCallable())
        {
            m_pCallFunction[what] = pIFirst->getAs<types::Callable>();        
        }
        else if (pIFirst != NULL && pIFirst->isString())
        {
            ConfigVariable::EntryPointStr* func = ConfigVariable::getEntryPoint(pIFirst->getAs<types::String>()->get(0));
            if (func == NULL)
            {
                sprintf(errorMsg,_("%ls: unable to find entry point %ls.\n"), m_wstrCaller.c_str(), pIFirst->getAs<types::String>()->get(0));
                throw ast::InternalError(errorMsg);
            }
            else if (pList->getSize()>2)
            {
                sprintf(errorMsg,_("%ls: only one parameter is allowed.\n"), m_wstrCaller.c_str(), pIFirst->getAs<types::String>()->get(0));
                throw ast::InternalError(errorMsg);                
            }
            else if (pList->get(1)->isDouble() == false || pList->get(1)->getAs<types::Double>()->isComplex())
            {
                sprintf(errorMsg,_("%ls: parameter must be a real matrix.\n"), m_wstrCaller.c_str(), pIFirst->getAs<types::String>()->get(0));
                throw ast::InternalError(errorMsg);                                
            }
            else
            {
                m_pEntryPointFunctionWithPar[what] = func->functionPtr;
                m_pParameters[what].push_back(pList->get(1));
            }            
        }
        else
        {
            sprintf(errorMsg,_("%ls: first element of parameter %s should be a function or a string.\n"), m_wstrCaller.c_str(), pstrArgName[what]);
            throw ast::InternalError(errorMsg);
        }
        for (int i=1; i<pList->getSize(); i++)
        {
            // mandatory IncreaseRef !
            pList->get(i)->IncreaseRef();
            m_pParameters[what].push_back(pList->get(i));
        }
    }
    else if (pIn->isString())
    {
        ConfigVariable::EntryPointStr* func = ConfigVariable::getEntryPoint(pIn->getAs<types::String>()->get(0));
        if (func == NULL)
        {
            sprintf(errorMsg,_("%ls: unable to find entry point %ls.\n"), m_wstrCaller.c_str(), pIn->getAs<types::String>()->get(0));
            throw ast::InternalError(errorMsg);
        }
        else
        {
            m_pEntryPointFunction[what] = func->functionPtr;
        }
    }
    else
    {
        sprintf(errorMsg, _("%ls: parameter %s should be a string, a function or a list.\n"), m_wstrCaller.c_str(), pstrArgName[what]);
        throw ast::InternalError(errorMsg);
    }

    if (m_pCallFunction[what] != NULL)
    {
        m_pCallFunctionName[what] = wide_string_to_UTF8(m_pCallFunction[what]->getName().c_str());        
    }
}

void OdeManager::parseMatrices(types::typed_list &in)
{
    char errorMsg[256];
    types::Double *pDblTSpan = NULL;

    // Parse second argument "tspan"
    if (in[1]->isDouble() && !in[1]->getAs<types::Double>()->isComplex())
    {
        pDblTSpan = in[1]->getAs<types::Double>();

        // Check if we are extending a previously computed solution
        if (m_prevManager != NULL)
        {
            if (pDblTSpan->getSize() > 1)
            {
                sprintf(errorMsg, _("%ls: Wrong size for input argument #%d: expecting %d.\n"), m_wstrCaller.c_str(), 1, 1);                
                throw ast::InternalError(errorMsg);
            }
            double dblTFinal = m_prevManager->m_dblVecTOut.back();
            if ( ((pDblTSpan->get(0)-dblTFinal)*(dblTFinal-m_prevManager->m_dblT0)) < 0)
            {
                sprintf(errorMsg, _("%ls: parameter #%d must be greater than %g.\n"), m_wstrCaller.c_str(), 2, dblTFinal);
                throw ast::InternalError(errorMsg);
            }
            // Simulate option t0 = tFinal
            m_dblOptT0 = dblTFinal;
        }

        m_pDblTSpan = pDblTSpan->clone();
        if (m_pDblTSpan->getSize() > 2)
        {
            for (int i=2; i<m_pDblTSpan->getSize(); i++)
            {
                if ( (m_pDblTSpan->get(i)-m_pDblTSpan->get(i-1)) * (m_pDblTSpan->get(i-1)-m_pDblTSpan->get(i-2)) < 0)
                {
                    sprintf(errorMsg, _("%ls: values in parameter #%d should be monotone.\n"), m_wstrCaller.c_str(), 2);
                    throw ast::InternalError(errorMsg);          
                }
            }
        }
    }
    else
    {
        sprintf(errorMsg, _("%ls: Wrong type for input argument #%d: A real matrix expected.\n"), m_wstrCaller.c_str(), 2);
        throw ast::InternalError(errorMsg);          
    }

    // Parse third argument "y0" 
    // y0 is parsed as an option when extending solution
    if (in.size() > 2)
    {
        if (in[2]->isDouble())
        {
            types::Double *pDbl =  in[2]->getAs<types::Double>();
            if (pDbl->isVector() && pDbl->getRows() == 1)
            {
                 pDbl->transpose((types::InternalType *&)m_pDblY0);
            }
            else
            {
                m_pDblY0 = pDbl->clone();                
            }
        }
        else
        {
            sprintf(errorMsg, _("%ls: Wrong type for input argument #%d: A real or complex matrix expected.\n"), m_wstrCaller.c_str(),3);
            throw ast::InternalError(errorMsg);          
        }
        m_odeIsComplex = m_pDblY0->isComplex(); 
    }
    else
    {
        // by default, extend solution with Y0 = final value of previous solution
        int iNbEq = m_prevManager->m_iNbEq;
        double *pdblPrevYFinal = NV_DATA_S(m_prevManager->m_N_VectorY);

        m_pDblY0 = m_prevManager->m_pDblY0->clone();
        m_pDblY0->setComplex(m_prevManager->m_odeIsComplex);
        std::copy(pdblPrevYFinal, pdblPrevYFinal+iNbEq, m_pDblY0->get());
        if (m_pDblY0->isComplex())
        {
            std::copy(pdblPrevYFinal+iNbEq, pdblPrevYFinal+2*iNbEq, m_pDblY0->getImg());
        }
        m_odeIsComplex = m_prevManager->m_odeIsComplex; 
    }
    m_iNbEq = m_pDblY0->getSize();
    m_iSizeOfCVODEInput[RHS] = m_iNbEq;
    m_iSizeOfCVODEInput[JACY] = m_iNbEq*m_iNbEq;        
}

void OdeManager::parseOptions(types::optional_list &opt)
{
    char errorMsg[256];
    
    int iMaxmethodOrder = m_wstrMethod == L"adams" ? 12 : 5;

    // check if called by cvode_extend or cvode_solve
    if (m_prevManager != NULL)
    {
        // y0 option 
        if (opt.find(L"y0") != opt.end())
        {
            if (opt[L"y0"]->isDouble())
            {
                types::Double *pDbl = opt[L"y0"]->getAs<types::Double>();
                bool bSameDims = true;
                if (pDbl->getDims() == m_pDblY0->getDims())
                {
                    for (int i=0; i<pDbl->getDims(); i++)
                    {
                        bSameDims &= (pDbl->getDimsArray()[i] == m_pDblY0->getDimsArray()[i]);
                    }
                }
                if (bSameDims == false)
                {
                    sprintf(errorMsg, _("%ls: Wrong size for option \"y0\": expecting %d x %d.\n"), m_wstrCaller.c_str(), m_pDblY0->getRows(), m_pDblY0->getCols());
                    throw ast::InternalError(errorMsg);
                }
                if (m_prevManager->m_odeIsComplex == false && pDbl->isComplex())
                {
                    sprintf(errorMsg, _("%ls: Wrong type (complex) for option \"y0\": extended ode is real.\n"), m_wstrCaller.c_str());
                    throw ast::InternalError(errorMsg);                                                    
                }
                m_pDblY0->killMe();
                m_pDblY0 = pDbl->clone();
                m_pDblY0->setComplex(m_odeIsComplex);
                
            }
            else
            {
                sprintf(errorMsg, _("%ls: Wrong type for option \"y0\": a double vector is expected.\n"), m_wstrCaller.c_str());
                throw ast::InternalError(errorMsg);                
            }
        }        

        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"rtol", &m_dblRtol, m_prevManager->m_dblRtol, {1e-15, 1});
        getDoubleVectorInPlist(m_wstrCaller.c_str(), opt, L"atol", m_dblVecAtol, m_prevManager->m_dblVecAtol, {1e-15, 1}, m_iNbEq);
        getStringInPlist(m_wstrCaller.c_str(), opt, L"method", m_wstrMethod, m_prevManager->m_wstrMethod, {L"adams",L"bdf"});
        getIntInPlist(m_wstrCaller.c_str(), opt, L"maxorder", &m_iMaxOrder, m_prevManager->m_iMaxOrder, {1, iMaxmethodOrder});
        getIntInPlist(m_wstrCaller.c_str(), opt, L"refine", &m_iNRefine, m_prevManager->m_iNRefine, {1, INT_MAX});
        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"h0", &m_dblInitialStep, m_prevManager->m_dblCurrentStep, {0, std::numeric_limits<double>::infinity()});        
        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"hmax", &m_dblMaxStep, m_prevManager->m_dblMaxStep, {0, std::numeric_limits<double>::infinity()});        
        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"hmin", &m_dblMinStep, m_prevManager->m_dblMinStep, {0, std::numeric_limits<double>::infinity()});        
        getIntVectorInPlist(m_wstrCaller.c_str(), opt, L"positive", m_iVecPositive, m_prevManager->m_iVecPositive, {1,m_iNbEq}, {1,m_iNbEq});        

        // By default we use rhs of previous solution
        parseFunction(opt.find(L"f") != opt.end() ? opt[L"f"] : NULL, RHS);
        
        computeFunction(m_dblT0, NULL, NULL, RHS, NULL);
        if (m_odeIsComplex && m_prevManager->m_odeIsComplex == false)
        {
            sprintf(errorMsg, _("%ls: Wrong outpout type (complex) for option \"f\": extended ode is real.\n"), m_wstrCaller.c_str());
            throw ast::InternalError(errorMsg);                                            
        }
        // By default we use Jacobian of previous solution
        parseFunction(opt.find(L"jacobian") != opt.end() ? opt[L"jacobian"] : NULL, JACY);
        // By default we use intermediate callback of previous solution
        parseFunction(opt.find(L"intcb") != opt.end() ? opt[L"intcb"] : NULL, JACY);
        
        // Inherit events function (cannot be changed from previous solution. TODO: if no events before, could be added in extension)
        parseFunction(NULL, EVENTS);
        setupEvents();

        if (m_odeIsComplex == true && m_pCallFunction[JACY] != NULL)
        {
            sprintf(errorMsg, _("%ls: jacobian + complex ODE is not supported.\n"), m_wstrCaller.c_str());
            throw ast::InternalError(errorMsg);
        }
    }
    else
    {
        // Determine if ode is complex (can be already detected if Y0 is complex)
        // in the case where RHS is a Scilab function
        if (m_odeIsComplex == false && m_pCallFunction[RHS] != NULL)
        {
            computeFunction(m_dblT0, NULL, NULL, RHS, NULL);
            if (m_odeIsComplex) // complexify Y0 (a clone of user Y0)
            {
                m_pDblY0->setComplex(true);
            }
        }
     
        // Initial time
        if (m_pDblTSpan->getSize() > 1 )
        {
            if (m_pDblTSpan->get(0) < m_pDblTSpan->get(1))
            {
                getDoubleInPlist(m_wstrCaller.c_str(), opt, L"t0", &m_dblOptT0, m_dblOptT0, {-std::numeric_limits<double>::infinity(), m_pDblTSpan->get(0)});                        
            }
            else
            {
                getDoubleInPlist(m_wstrCaller.c_str(), opt, L"t0", &m_dblOptT0, m_dblOptT0, {m_pDblTSpan->get(0), std::numeric_limits<double>::infinity()});                        
            }
        }
        else // m_TSpan->getSize() == 1
        {
            getDoubleInPlist(m_wstrCaller.c_str(), opt, L"t0", &m_dblOptT0, m_dblOptT0, {});
        }

        // method
        int iMaxmethodOrder = m_wstrMethod == L"adams" ? 12 : 5;

        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"rtol", &m_dblRtol, 1e-4, {1e-15, 1});
        getDoubleVectorInPlist(m_wstrCaller.c_str(), opt, L"atol", m_dblVecAtol, {1e-6}, {1e-15, 1}, m_iNbEq);        
        getStringInPlist(m_wstrCaller.c_str(), opt, L"method", m_wstrMethod, L"adams", {L"adams",L"bdf"});
        getIntInPlist(m_wstrCaller.c_str(), opt, L"maxorder", &m_iMaxOrder, iMaxmethodOrder, {1, iMaxmethodOrder});
        getIntInPlist(m_wstrCaller.c_str(), opt, L"refine", &m_iNRefine, 1, {1, INT_MAX});
        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"h0", &m_dblInitialStep, 0, {0, std::numeric_limits<double>::infinity()});        
        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"hmax", &m_dblMaxStep, 0, {0, std::numeric_limits<double>::infinity()});        
        getDoubleInPlist(m_wstrCaller.c_str(), opt, L"hmin", &m_dblMinStep, 0, {0, std::numeric_limits<double>::infinity()});        
        getIntVectorInPlist(m_wstrCaller.c_str(), opt, L"positive", m_iVecPositive, {}, {1,m_iNbEq}, {1,m_iNbEq});        

        // Jacobian
        if (opt.find(L"jacobian") != opt.end())
        {
            if (m_odeIsComplex == true)
            {
                sprintf(errorMsg, _("%ls: jacobian + complex ODE is not supported.\n"), m_wstrCaller.c_str());
                throw ast::InternalError(errorMsg);                
            }
            parseFunction(opt[L"jacobian"], JACY);
        }

        // Events    
        if (opt.find(L"events") != opt.end())
        {
            parseFunction(opt[L"events"], EVENTS);
            setupEvents();
        }        

        // Intermediate callback
        if (opt.find(L"intcb") != opt.end())
        {
            parseFunction(opt[L"intcb"], INTCB);
        }
    }
}