//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2021 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v3.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "OdeManager.hxx"
#include "CVODEBridge.hxx"

void OdeManager::solve()
{
    CVodeMem cv_mem;
    int iFlag;
    int iLastOrder;
    int iNbErrors = 0; 
    int iNbSteps = 0;
    int iFirstStep = 0;
    double dblTime;
    double dblCurrTime;
    double dblStep;
    double dblFinalTime;
    bool bTerminalEvent = false;
    char errorMsg[256];

    // std::list<std::vector<double>> pDblYEventList;
    // std::list<std::vector<double>> pDblIndexEventList;
    // std::vector<double> eventTimeList;
        
    cv_mem = (CVodeMem) m_cvode_mem;

    CVODEIntCb(m_dblT0, m_N_VectorY, L"init", (void *)this);
 
    if ((m_pDblTSpan->getSize() == 2 && std::isnan(m_dblOptT0)) || m_iRetCount <= 1)        
    {
        /*
            Case where we compute solution only at method steps guided only by precision (RTOL, ATOL) requirements.
            This occurs when tspan is of the kind [t0 t1] or in the case of sol = method() syntax, i.e. when we yield
            a solution structure with a pointer to the OdeManager instance for further custom evaluation by interpolation 
            or time extension.
        */
        if (CVodeSetStopTime(m_cvode_mem, m_pDblTSpan->get(m_pDblTSpan->getSize()-1)) != CV_SUCCESS)
        {
            sprintf(errorMsg,"CVodeSetStopTime error\n");             
            throw ast::InternalError(errorMsg);
        }
        std::list<std::vector<double>> pDblDerOutList;
        dblFinalTime = m_pDblTSpan->get(m_pDblTSpan->getSize()-1);

        // Copy (t0,y0) only if not extending a previous solution
        if (m_prevManager == NULL)
        {

            // TODO: reserve space for vectors
            m_indexNS.push_back(0);
            m_vecYOut.push_back(std::vector<double>(NV_DATA_S(m_N_VectorY),NV_DATA_S(m_N_VectorY) + m_iNbRealEq));
            m_dblVecTOut.push_back(m_dblT0);
        }
        else
        {
            // new values will be appended to previous vectors
            // TODO: reserve space for vectors
            m_vecYOut = m_prevManager->m_vecYOut;
            m_dblVecTOut = m_prevManager->m_dblVecTOut; // Beware, m_dblVecTOut has T0 hence its size is m_iVecOrder.size()+1
            m_dblVecCurrTime = m_prevManager->m_dblVecCurrTime;
            m_iVecOrder = m_prevManager->m_iVecOrder;
            m_indexNS = m_prevManager->m_indexNS;
            // events
            m_dblVecYEvent = m_prevManager->m_dblVecYEvent;
            m_dblVecIndexEvent = m_prevManager->m_dblVecIndexEvent;
            m_dblVecEventTime = m_prevManager->m_dblVecEventTime;
            // index of further steps takes into account previous values
            iFirstStep = m_prevManager->m_iVecOrder.size();
        }        

        iNbSteps = iFirstStep;

        for (; m_dblT0 != dblFinalTime;)
        {
            // CVode internal step
            iFlag = CVode(m_cvode_mem, dblFinalTime, m_N_VectorY, &dblTime, CV_ONE_STEP);
            if (iFlag == CV_SUCCESS || iFlag == CV_TSTOP_RETURN || iFlag == CV_ROOT_RETURN)
            {

                // Should not be necessary
                if (dblTime == m_dblVecTOut.back())
                {
                    CVODEErrHandler(CV_WARNING, NULL, NULL, NULL, (void *)this);
                    break;
                }
                
                std::vector<double> newY (NV_DATA_S(m_N_VectorY), NV_DATA_S(m_N_VectorY) + m_iNbRealEq);
                CVodeGetLastOrder(m_cvode_mem, &iLastOrder);
                CVodeGetCurrentTime(m_cvode_mem, &dblCurrTime);
                // Refine by interpolation if requested and [t,y] output 
                if (m_iNRefine > 1 && m_iRetCount > 1)
                {
                    N_Vector N_VectorYRefine = N_VNew_Serial(m_iNbRealEq);
                    double dblStepRef = (dblTime-m_dblVecTOut.back())/(double)m_iNRefine;
                    double dblRefTime = m_dblVecTOut.back()+dblStepRef;
                    for (int i=1; i < m_iNRefine; i++)
                    {
                        CVodeGetDky(m_cvode_mem, dblRefTime, 0, N_VectorYRefine);
                        std::vector<double> newY (NV_DATA_S(N_VectorYRefine), NV_DATA_S(N_VectorYRefine) + m_iNbRealEq);
                        m_vecYOut.push_back(newY);
                        m_dblVecTOut.push_back(dblRefTime);
                        dblRefTime +=dblStepRef;
                    }
                    N_VDestroy_Serial(N_VectorYRefine);
                }

                if (iFlag == CV_ROOT_RETURN)
                {                    
                    int *piRootsFound = new int(m_iNbEvents);
                    CVodeGetRootInfo(m_cvode_mem, piRootsFound);
                    std::vector<double> newIndex (piRootsFound, piRootsFound + m_iNbEvents);
                    delete piRootsFound;
                    
                    // Check if one event is a terminal event
                    if (m_iVecEventIsTerminal.size() > 0)
                    {
                        for (int i=0; i < m_iNbEvents; i++)
                        {
                            bTerminalEvent |= (newIndex[i] !=0 && m_iVecEventIsTerminal[i] == 1);
                        }
                        // bTerminalEvent == true will stop integration 
                    }
                    m_dblVecEventTime.push_back(dblTime);
                    m_dblVecYEvent.push_back(newY);
                    m_dblVecIndexEvent.push_back(newIndex);
                }
                m_vecYOut.push_back(newY);
                m_dblVecCurrTime.push_back(dblCurrTime);
                m_dblVecTOut.push_back(dblTime);
                m_iVecOrder.push_back(iLastOrder);
                iNbSteps++;
                if (m_iRetCount <= 1)
                { 
                    // Solution structure output
                    // Store current ordrer and Nordsieck History Array
                    // N_Vector cv_zn[L_MAX];   Nordsieck array, of size N x (q+1).
                    // zn[j] is a vector of length N (j=0,...,q)
                    // zn[j] = [1/factorial(j)] * h^j *(jth derivative of the interpolating polynomial)             
                    // indexNS is the cumulative index later used for solution interpolation
                    // indexNS[i] is the column index in whole matrix where NS matrix of timestep i starts.
                    m_indexNS.push_back(m_indexNS.back()+iLastOrder+1);

                    for (int i=0; i<iLastOrder+1; i++)
                    {
                        std::vector<double> vdblNordsieckVector (NV_DATA_S(cv_mem->cv_zn[i]), NV_DATA_S(cv_mem->cv_zn[i]) + m_iNbRealEq);
                        pDblDerOutList.push_back(vdblNordsieckVector);           
                    }                    
                }
                if (CVODEIntCb(dblTime, m_N_VectorY, L"step", (void *)this))                    
                {
                    break;
                }
                if (bTerminalEvent == true || iFlag == CV_TSTOP_RETURN)
                {
                    break;
                }
            }
            else
            {
                // Errors/Warnings not trapped by CVODEErrorFunc
                CVODEErrHandler(iFlag, NULL, NULL, NULL, (void *)this);
            }
        }

        if (m_iRetCount == 1) // Solution structure output, store Nordsieck matrices in m_pDblNSMatrices
        {
            double *pdblNS;

            m_pDblNSMatrices = new types::Double(m_iNbRealEq, m_indexNS.back());
            // Check if we are extending a previous solution
            if (m_prevManager != NULL)
            {
                types::Double *pDblPrevNSMatrices = m_prevManager->m_pDblNSMatrices;

                // copy previous NS matrices
                std::copy(pDblPrevNSMatrices->get(), pDblPrevNSMatrices->get()+pDblPrevNSMatrices->getSize(), m_pDblNSMatrices->get());
                pdblNS = m_pDblNSMatrices->get() + pDblPrevNSMatrices->getSize();
            }
            else
            {
                pdblNS = m_pDblNSMatrices->get();
            }
            int k=0;

            for (int i = iFirstStep; i < iNbSteps; i++)
            {
                for (int j = 0; j < m_iVecOrder[i]+1; j++)
                {
                    double *pdbl = pDblDerOutList.front().data();
                    
                    std::copy(pdbl, pdbl+m_iNbRealEq, pdblNS);
                    
                    pDblDerOutList.pop_front();
                    pdblNS += m_iNbRealEq;
                    k++;
                }
            }            
        }
    }
    else // Case where we compute solution at prescribed user timesteps, with method stepping and dense output interpolation
    {
        m_pDblYOut = createYOut(m_iNbEq, m_pDblTSpan->getSize());
        m_pDblTOut = m_pDblTSpan->clone();
        double *pdblYout = m_pDblYOut->get();
        double *pdblYoutImg = m_pDblYOut->getImg();

        if (m_dblT0 == m_pDblTSpan->get(0))
        {
            std::copy(m_pDblY0->get(),m_pDblY0->get()+m_iNbEq, pdblYout);
            pdblYout += m_iNbEq;
            if (m_odeIsComplex)
            {
                std::copy(m_pDblY0->getImg(),m_pDblY0->getImg()+m_iNbEq, pdblYoutImg);                
                pdblYoutImg += m_iNbEq;
            }
        }

        for (int i = (int) m_dblT0 == m_pDblTSpan->get(0); i<m_pDblTSpan->getSize();)
        {
            iFlag = CVode(m_cvode_mem, m_pDblTSpan->get(i), m_N_VectorY, &dblTime, CV_NORMAL);

            if (iFlag == CV_ROOT_RETURN)
            {
                std::vector<double> newY (NV_DATA_S(m_N_VectorY), NV_DATA_S(m_N_VectorY) + m_iNbRealEq);
                int *piRootsFound = new int(m_iNbEvents);
                CVodeGetRootInfo(m_cvode_mem, piRootsFound);
                std::vector<double> newIndex (piRootsFound, piRootsFound + m_iNbEvents);
                delete piRootsFound;

                m_dblVecEventTime.push_back(dblTime);
                m_dblVecYEvent.push_back(newY);
                m_dblVecIndexEvent.push_back(newIndex);

                // Check if one event is a terminal event
                if (m_iVecEventIsTerminal.size() > 0)
                {
                    for (int j=0; j < m_iNbEvents; j++)
                    {
                        bTerminalEvent |= (newIndex[j] !=0 && m_iVecEventIsTerminal[j] == 1);
                    }
                    if (bTerminalEvent == true)
                    {
                        // Stop integration at time of event, precisely (don't finish at next user step)
                        m_pDblTOut->set(i,dblTime);
                        std::copy(NV_DATA_S(m_N_VectorY),NV_DATA_S(m_N_VectorY)+m_iNbEq,pdblYout);
                        pdblYout += m_iNbEq;
                        if (m_odeIsComplex)
                        {
                            std::copy(NV_DATA_S(m_N_VectorY)+m_iNbEq,NV_DATA_S(m_N_VectorY)+2*m_iNbEq,pdblYout);
                            pdblYoutImg += m_iNbEq;
                        }
                        i++;
                        // do a resize of pdblYout and m_pDblTOut
                        m_pDblTOut->resize(1,i);
                        int iDims =  m_pDblYOut->getDims();
                        int *piDimsArray = m_pDblYOut->getDimsArray();
                        int *piNewDimsArray = new int(iDims);
                        std::copy(piDimsArray, piDimsArray+iDims, piNewDimsArray);
                        piNewDimsArray[iDims-1] = i;
                        m_pDblYOut->resize(piNewDimsArray, iDims);
                        delete piNewDimsArray;
                        break;
                    }
                }
            }
            else if (iFlag == CV_SUCCESS)
            {
                std::copy(NV_DATA_S(m_N_VectorY),NV_DATA_S(m_N_VectorY)+m_iNbEq,pdblYout);
                pdblYout += m_iNbEq;
                if (m_odeIsComplex)
                {
                    std::copy(NV_DATA_S(m_N_VectorY)+m_iNbEq,NV_DATA_S(m_N_VectorY)+2*m_iNbEq,pdblYoutImg);
                    pdblYoutImg += m_iNbEq;
                }
                i++;
            }
            else
            {
                // Errors/Warnings not trapped by CVODEErrorFunc
                CVODEErrHandler(iFlag, NULL, NULL, NULL, (void *)this);
            }
            if (CVODEIntCb(dblTime, m_N_VectorY, L"step", (void *)this))                    
            {
                break;
            }
        }
    }
    
    CVODEIntCb(dblTime, m_N_VectorY, L"done", (void *)this);

    // save last current step for future extension
    CVodeGetCurrentStep(m_cvode_mem, &m_dblCurrentStep);
}